#include <tap.h>
#include "libFireBird.h"

TAP_ID( 0x810714ce );
#define TAPNAME "CrashTrace"
TAP_PROGRAM_NAME(TAPNAME " v0.4");
TAP_AUTHOR_NAME("R2-D2");
TAP_DESCRIPTION("Enables a more detailed back trace on the serial port");
TAP_ETCINFO(__DATE__);

typedef dword (*FUNC)();

#define RAMSTART  (0x80000000)
#define RAMEND    (RAMSTART+(64*1024*1024))

#define EX_STK    (124)
#define EX_RA     (104)

dword *HookLocn, HookVal;
dword FWStart=RAMSTART, FWEnd, EndOfInt;
dword *TaskTbl, *CurrTask, *Tick;
tTAPTable *TAPTbl;
// MemoryMonitor(), getNowTimestamp(), SER_PrintfA(fmt, args...), DumpMem(addr,size,showhdr)
FUNC fwMemMon, fwGetNow, fwPrintF, fwDumpMem, fwMemInfo;

#define NUMREGS   (TOPPOS+1)
#define S0REG     (0x10)
#define S7REG     (0x17)
#define FPREG     (0x1e)
#define RAREG     (0x1f)
#define STKPOS    (0x20)
#define TOPPOS    (0x21)
#define VALIDPTR(x) ({dword _p=(dword)(x); _p >= RAMSTART && _p < RAMEND && (_p&0x3) == 0;})

short trackBack(dword ra, dword sp, dword *regs, byte *hasReg)
{
  dword *code = (dword*)ra;
  short spoff = 0;
  for( ; !spoff; code--)
  {
    dword inst = (*code) >> 16;
    word off = (*code) & 0xffff;
    if(inst == 0x27bd) // addiu $sp,X
      spoff = (short)off;
    else if((inst & 0xafa0) == 0xafa0) // sw $reg,X($sp)
    {
      dword reg = inst & 0x1f;
      dword p = sp+off;
      if(VALIDPTR(p))
      {
        hasReg[reg] = 1;
        regs[reg] = *(dword*)p;
      }
    }
  }
  if(spoff<0) // Proper (-ve) stack adj found
    regs[TOPPOS] = (dword)(code+1);
  else // End of function (+ve) stack adj found, so this used none
  {
    int i;
    for(i=0; i<NUMREGS-2; i++)
      hasReg[i] = 0;
    regs[TOPPOS] = ra;
    spoff = 0;
  }
  regs[STKPOS] = sp-spoff;
  return spoff;
}

#define NUMCALLS  (11)

void fillTrackBack(dword ra, dword altRA, dword sp, dword base, dword (*regs)[NUMREGS], byte (*hasReg)[NUMREGS])
{
  dword top = 0;
  short spoff;
  int i;
  memset(hasReg, 0, NUMCALLS*NUMREGS);
  memset(regs, 0, NUMCALLS*NUMREGS*sizeof(dword));
  regs[0][RAREG] = ra;
  regs[1][RAREG] = altRA;
  regs[0][STKPOS] = sp;
  for(i=1; i<NUMCALLS && VALIDPTR(sp) && VALIDPTR(ra) && EndOfInt<=ra && (top<base || top>base+20); i++)
  {
    spoff = trackBack(ra, sp, regs[i], hasReg[i]);
//fwPrintF("%d. task=%d base=%p code=%p top=%p spoff=%d\n", i, *CurrTask, base, ra, top, spoff);
    if(hasReg[i][RAREG])
      ra = regs[i][RAREG];
    else if(i==1)
      ra = altRA;
    else
      ra = 0;
    sp -= spoff;
    top = regs[i][TOPPOS];
  }
}

#define RNDMASK (0xFFFFFFF0)

void dumpMem(dword addr, char *type)
{
  dword a=(addr-0x80)&RNDMASK;
  if(VALIDPTR(a))
  {
    fwPrintF("\nDumping bytes from %08x for %s (%08x)\n", a, type, addr);
    while(a < addr+0x80)
    {
      int i;
      dword *p = (dword*)a;
      fwPrintF("%08x: ", p);
      for(i=0; i<4; i++, p++)
        fwPrintF(addr==(dword)p ? "*%08x*" : " %08x ", *p);
      fwPrintF("\n");
      a = (dword)p;
    }
    //fwDumpMem(a, 0x100, FALSE);
  }
  else
    fwPrintF("\nBad %s (%08x) - no dump\n", type, addr);
}

#define COLS (6)

// MemoryMonitor() call in DumpException
void hook(dword slot)
{
  register dword _s6 asm("s6");
  dword epc=_s6;
  register dword _sp asm("sp"), _ra asm("ra");
  dword sp=_sp, ra=_ra;
  dword c, now = fwGetNow(), t = *Tick;
  dword base = TaskTbl[*CurrTask];
  static byte hasReg[NUMCALLS][NUMREGS];
  static dword regs[NUMCALLS][NUMREGS];
  dumpMem(epc, "EPC");
  fillTrackBack(ra, ra, sp, base, regs, hasReg);
/*
  fwPrintF(" ra=[%p %p %p %p] epc=%p\nstk=[%p %p %p %p]\n",
    regs[1][TOPPOS], regs[2][TOPPOS], regs[3][TOPPOS], regs[4][TOPPOS], epc,
    regs[1][STKPOS], regs[2][STKPOS], regs[3][STKPOS], regs[4][STKPOS]);
*/

  sp = regs[2][STKPOS]+EX_STK;  // exception_handler stack adjs
  ra = *(dword*)(sp-EX_STK+EX_RA); // $ra at exception point
  dumpMem(sp, "SP @ EPC fn");
  dumpMem(ra, "RA");

  fillTrackBack(epc, ra, sp, base, regs, hasReg);
  dumpMem(regs[1][STKPOS], "SP @ RA fn");

  fwPrintF("\n** Stack back trace @ %02d:%02d (uptime: %dd %dh %dm %ds) **\n",
    (now>>8)&0xff, now&0xff, (t/(24*60*60*100)), (t/(60*60*100))%24, (t/(60*100))%60, (t/100)%60);
  for(c=1; c<NUMCALLS && regs[c][TOPPOS] != 0; c++)
  {
    dword i, n=0, *r=regs[c], fn=regs[c][TOPPOS], ra=regs[c-1][RAREG];
    fwPrintF("%d) fn=%08x ra=%08x sp=%08x", c, fn, ra, regs[c-1][STKPOS]);
    if(fn == base)
      fwPrintF(" in TASK ROOT *end of trace*");
    else if(fn >= FWStart && fn < FWEnd)
      fwPrintF(" in FIRMWARE CODE");
    else if(TAPTbl == NULL)
      fwPrintF(" in *UNKNOWN*");
    else
    {
      for(i = 0; i < TAP_MAX; i++)
      {
        tTAPTable *t = &TAPTbl[i];
        if(VALIDPTR(t->LoadAddress) && fn >= t->LoadAddress)
        {
          tTAPEntryPoints *e = t->EntryPoints;
          if(VALIDPTR(e) && fn < t->LoadAddress+e->_tap_textEnd)
          {
            tTAPHeader *h = t->TAPHeader;
            fwPrintF(" in TAP @ +%08x %.20s", ra - t->LoadAddress, VALIDPTR(h) && VALIDPTR(h->Name) ? h->Name : "*CORRUPT*");
            break;
          }
        }
      }
      if(i == TAP_MAX)
        fwPrintF(" **IN WEEDS**");
    }
    for(i=S0REG; i<=S7REG; i++)
    {
      if(hasReg[c][i])
      {
        if((n++%COLS) == 0)
          fwPrintF("\n    ");
        fwPrintF(" s%d=%08x", i-S0REG, r[i]);
      }
    }
    if(hasReg[c][FPREG])
    {
      if((n++%COLS) == 0)
        fwPrintF("\n    ");
      fwPrintF(" fp=%08x", r[FPREG]);
    }
    fwPrintF("\n");
  }
  fwPrintF("\n");
  if(TAPTbl)
  {
    int i;
    fwPrintF("** TAP list **\n");
    for(i = 0; i < TAP_MAX; i++)
    {
      tTAPTable *t = &TAPTbl[i];
      if(t->LoadAddress)
      {
        tTAPEntryPoints *e = t->EntryPoints;
        tTAPHeader *h = t->TAPHeader;
        if(VALIDPTR(t->LoadAddress) && VALIDPTR(e) && VALIDPTR(h))
          fwPrintF("%2d] %08x -> %08x (%08x) %.20s\n",
            i, t->LoadAddress, t->LoadAddress+e->_tap_textEnd, t->LoadAddress+e->_end, h->Name);
        else
          fwPrintF("%2d] %08x *CORRUPT TAP* (entry=%08x, hdr=%08x)\n",
            i, t->LoadAddress, e, h);
      }
    }
    fwPrintF("\n");
  }
  if(fwMemInfo)
  {
    dword total, free, avail;
    fwMemInfo(&total, &free, &avail);
    fwPrintF("** Memory **\nfree=%2d,%03d,%03d avail=%2d,%03d,%03d\n\n",
      free/(1000*1000), (free/1000)%1000, free%1000, avail/(1000*1000), (avail/1000)%1000, avail%1000);
  }
  fwMemMon();
}

void addHook()
{
  if(*HookLocn == HookVal)
  {
    *HookLocn = JAL_CMD | REL_ADDR(hook);
    TAP_Print(TAPNAME ": hook added @ %08x\n", TAP_GetTick());
  }
}

void removeHook()
{
  if(*HookLocn == (JAL_CMD | REL_ADDR(hook)))
    *HookLocn = HookVal;
}

dword TAP_EventHandler( word event, dword param1, dword param2 )
{
  return param1;
}

void *FindFnStart(void *vp)
{
  dword *p = vp;
  while((*p & 0xfffff000) != 0x27bdf000 && (*p & CMD_MASK) != JMP_CMD)
    p--;
  return p;
}

int TAP_Main(void)
{
  // exception_handler (end) @ 800002dc, stk adj: -124, ra offset: 104
  dword ex = FindInstructionSequence(
    "8fbf0068 8fbe0064 8fb70060 8fb6005c 8fb50058 8fb40054 8fb30050 8fb2004c 8fb10048 8fb00044 8fb90040 8fb8003c 8faf0038 8fae0034 8fad0030 8fac002c 8fab0028 8faa0024 8fa90020 8fa8001c 8fa70018 8fa60014 8fa50010 8fa4000c 8fa30008 8fa20004 8fa10000 27bd007c 42000018",
    "ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff",
    0x80000180, 0x80001000, 0, FALSE);
  // DumpException @ 80001b18, EPC code using $s6
  dword ep = FindInstructionSequence(
    "3c04802d 24842520 02c02825 0c02b5b8 00000000 3c188000 02d8082b 1420000f 00000000 3c19c000 02d9082b 1020000b 00000000",
    "ffff0000 ffff0000 ffffffff fc000000 ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff ffffffff",
    0x80001800, 0x8000a000, 0, FALSE);
  // DumpException @ 80001c04, MemoryMonitor call @+8
  dword *mm = (void*)FindInstructionSequence(
    "3c04802d 248425a0 03c02825 0c02b5b8 00000000 27de0001 1000ffef 00000000 0c02ad7a 00000000",
    "ffff0000 ffff0000 ffffffff fc000000 ffffffff ffffffff ffffffff ffffffff fc000000 ffffffff",
    0x80001800, 0x8000a000, 0, FALSE);
  // InitTasks @ 8000069c, Tick: gp+short@+1, TaskFnTable: LUI@+9 + short@+11
  word *it = (void*)FindInstructionSequence(
    "af8095ac 8f998010 27390001 af998010 3c02802d 8c42233c",
    "ffff0000 ffff0000 ffffffff ffff0000 ffff0000 ffff0000",
    0x80000600, 0x80001000, 0, FALSE);
  // TaskSwitch @ 800006e4, CurrTask: gp+short@+7
  short *ts = (void*)FindInstructionSequence(
    "af80957c 3c1981da 2739b7e0 8f989580 0018c080 0338c821 af3e0000",
    "ffff0000 ffe00000 fc000000 ffe00000 ffe007ff fc0007ff fc1fffff",
    0x80000600, 0x80001000, 0, FALSE);
  // DumpMem(addr,size,showhdr) @ 801147c0, SER_PrintfA call @+6
  dword *dm = (void*)FindInstructionSequence(
    "00808825 10c00006 00a09025 02202825 3c048025 24847cc0 0c02b5b8 02403025 1240003f 0000b025",
    "ffff07ff ffffffff ffff07ff fc1fffff ffff0000 ffff0000 fc000000 fc1fffff fc1fff00 ffff07ff",
    0x80100000, 0x80160000, 0, FALSE);
  // a key handler @ 80169684, subTimestamps call@+1, addTimestamps call@+5, getNowTimestamp call@+8
  dword *gn = (void*)FindInstructionSequence(
    "03c02025 0c04c5cd 24050c00 10000004 0040f025 0c04c56f 24050c00 0040f025 0c04f110 00000000",
    "fc1fffff fc000000 ffffffff ffffffff ffff07ff fc000000 ffffffff ffff07ff fc000000 ffffffff",
    0x80150000, 0x801b0000, 0, FALSE);
  // fini @ 802566c8 (Firmware End)
  dword fn = FindInstructionSequence(
    "27bdfff8 afbf0000 8fbf0000 27bd0008 03e00008 00000000",
    "FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFF",
    0x801A0000, 0x80320000, 0, FALSE);
  // TerminateTAP @ 801787d0, tap_table: LUI@+1 + short@+3
  word *tt = (void*)FindInstructionSequence(
    "3c168018 26d69508 000f7940 02cf7821 8de40008 10800004 00002825",
    "ffff0000 ffff0000 ffffffff ffff07ff fc1fffff ffffffff ffffffff",
    0x80160000, 0x801c0000, 0, FALSE);
  // MemInfo(&total,&free,&avail) @ 80002560
  fwMemInfo = (void*)FindInstructionSequence(
    "00008025 02008825 0c0000f9 00000000 0040a025 979395b0 3273ffff 3419ffff 12790033 00000000",
    "ffffffff ffffffff fc000000 ffffffff ffffffff ffff0000 ffffffff ffffffff ffffffff ffffffff",
    0x80001800, 0x80006000, 0, TRUE);
  InitTAPex();
  if(!ex || !ep || !mm || !it || !ts || !dm || !gn || !fn)
  {
    char str[80];
    TAP_SPrint(str, "SysID=%d ApplID=%x", SysID, ApplID);
    TAP_Print(TAPNAME ": %s\n", str);
    ShowMessageWin(TAPNAME, "Unsupported firmware", str, 500);
    return 0;
  }
  EndOfInt = ex+120;
  HookLocn = mm+8;
  HookVal = *HookLocn;
  fwMemMon = (void*)ABS_ADDR(HookVal);
  Tick = (void*)(FWgp+(short)(it[1]));
  TaskTbl = (void*)((it[9]<<16)+(short)(it[11]));
  CurrTask = (void*)(FWgp+ts[7]);
  fwDumpMem = FindFnStart(dm);
  fwPrintF = (void*)ABS_ADDR(dm[6]);
  fwGetNow = (void*)ABS_ADDR(gn[8]);
  FWEnd = fn+24;
  if(tt == NULL)
    TAPTbl = NULL;
  else
    TAPTbl = (void*)((tt[1]<<16)+(short)(tt[3]));
  addHook();
  return 1;
/*
  TAP_Print("end=%p fwend=%p hook=%p val=%p task=%p curr=%p\n",
    EndOfInt, FWEnd, HookLocn, HookVal, TaskTbl, CurrTask);
  TAP_Print("tick=%p mm=%p dump=%p printf=%p getnow=%p tap=%p\n",
    Tick, fwMemMon, fwDumpMem, fwPrintF, fwGetNow, TAPTbl);
  return 0;
*/
}
